@@ -25,9 +25,12 @@ extensions = ['images',
 
 if sphinxcontrib:
     extensions.append('sphinxcontrib.spelling')
-extlinks = {'doi': ('https://doi.org/%s', 'doi: %s'),
-            'arxiv': ('https://arxiv.org/abs/%s', 'arXiv: %s'),
-            'xkcd': ('https://xkcd.com/%s', 'XKCD: %s')}
+extlinks = {
+    'doi': ('https://doi.org/%s', 'doi: %s'),
+    'arxiv': ('https://arxiv.org/abs/%s', 'arXiv: %s'),
+    'mr': ('https://gitlab.com/gpaw/gpaw/-/merge_requests/%s', 'MR: !%s'),
+    'issue': ('https://gitlab.com/gpaw/gpaw/-/issues/%s', 'issue: #%s'),
+    'xkcd': ('https://xkcd.com/%s', 'XKCD: %s')}
 spelling_word_list_filename = 'words.txt'
 spelling_show_suggestions = True
 templates_path = ['templates']
@@ -45,6 +48,7 @@ intersphinx_mapping = {
     'python': ('https://docs.python.org/3.10', None),
     'ase': ('https://wiki.fysik.dtu.dk/ase', None),
     'numpy': ('https://numpy.org/doc/stable', None),
+    'cupy': ('https://docs.cupy.dev/en/stable', None),
     'scipy': ('https://docs.scipy.org/doc/scipy', None),
     'pytest': ('https://docs.pytest.org/en/stable', None),
     'mayavi': ('http://docs.enthought.com/mayavi/mayavi', None)}
@@ -57,6 +61,7 @@ nitpick_ignore = [('py:class', 'gpaw.calculator.GPAW'),
                   ('py:class', 'ase.atoms.Atoms'),
                   ('py:class', 'gpaw.point_groups.group.PointGroup'),
                   ('py:class', 'UniformGridFunctions'),
+                  ('py:class', 'gpaw.core.arrays.DomainType'),
                   ('py:class', 'DomainType'),
                   ('py:class', 'Path'),
                   ('py:class', 'Vector'),
@@ -66,7 +71,10 @@ nitpick_ignore = [('py:class', 'gpaw.calculator.GPAW'),
                   ('py:class', 'Array2D'),
                   ('py:class', 'Array3D'),
                   ('py:class', 'MPIComm'),
-                  ('py:class', 'DomainType'),
+                  ('py:class', 'numpy._typing._dtype_like._SupportsDType'),
+                  ('py:class', 'numpy._typing._dtype_like._DTypeDict'),
+                  ('py:class', 'DTypeLike'),
+                  ('py:class', 'ModuleType'),
                   ('py:class', 'IO')]
 
 html_theme = 'sphinx_rtd_theme'
