@@ -8,7 +8,10 @@ calc = GPAW(mode='fd',
             h=0.2,
             xc='PBE',
             nbands=16,
-            poissonsolver={'remove_moment': 1 + 3 + 5},
+            poissonsolver={
+                'name': 'MomentCorrectionPoissonSolver',
+                'poissonsolver': 'fast',
+                'moment_corrections': 1 + 3 + 5},
             convergence={'density': 1e-12},
             txt='gs.out')
 atoms.calc = calc
