@@ -27,7 +27,8 @@ def do(freq):
     dmax = max(d_yx.min(), d_yx.max())
     vmax = 0.9 * dmax
     vmin = -vmax
-    plt.pcolormesh(X, Y, d_yx.T, cmap='RdBu_r', vmin=vmin, vmax=vmax)
+    plt.pcolormesh(X, Y, d_yx.T, cmap='RdBu_r', vmin=vmin, vmax=vmax,
+                   shading='auto')
     contours = np.sort(np.outer([-1, 1], [0.02]).ravel() * dmax)
     plt.contour(X, Y, d_yx.T, contours, cmap='RdBu_r', vmin=-1e-10, vmax=1e-10)
     for atom in atoms:
