@@ -11,11 +11,13 @@ with open(f'results-{h:.2f}.txt', 'w') as resultfile:
         system = molecule(name)
         system.set_cell((a, a, a))
         system.center()
-    
-        calc = GPAW(h=h,
+
+        if name in ['H', 'O']:
+            hund = True
+        else:
+            hund = False
+        calc = GPAW(h=h, hund=hund,
                     txt=f'gpaw-{name}-{h:.2f}.txt')
-        if name == 'H' or name == 'O':
-            calc.set(hund=True)
     
         system.calc = calc
     
