@@ -7,17 +7,17 @@ import contextlib
 from pathlib import Path
 from typing import List, Dict, Union, Any, TYPE_CHECKING
 
-__version__ = '22.8.0'
+__version__ = '23.6.0'
 __ase_version_required__ = '3.22.1'
 __all__ = ['GPAW',
            'Mixer', 'MixerSum', 'MixerDif', 'MixerSum2',
+           'MixerFull',
            'CG', 'Davidson', 'RMMDIIS', 'DirectLCAO',
            'PoissonSolver',
            'FermiDirac', 'MethfesselPaxton', 'MarzariVanderbilt',
            'PW', 'LCAO', 'FD',
            'restart']
 
-
 setup_paths: List[Union[str, Path]] = []
 is_gpaw_python = '_gpaw' in sys.builtin_module_names
 dry_run = 0
@@ -64,8 +64,9 @@ with broadcast_imports:
         sys.setdlopenflags(old_dlopen_flags)
     import _gpaw
 
+SCIPY_VERSION = [int(x) for x in scipy.__version__.split('.')[:2]]
 
-if getattr(_gpaw, 'version', 0) < 1:
+if getattr(_gpaw, 'version', 0) != 3:
     raise ImportError('Please recompile GPAW''s C-extensions!')
 
 
@@ -171,7 +172,7 @@ if debug:
         try:
             a.fill(np.nan)
         except ValueError:
-            a.fill(-1000000)
+            a.fill(42)
         return a
 
     def empty_like(*args, **kwargs):
@@ -179,7 +180,7 @@ if debug:
         try:
             a.fill(np.nan)
         except ValueError:
-            a.fill(-2000000)
+            a.fill(-42)
         return a
 
     np.empty = empty
@@ -188,7 +189,8 @@ if debug:
 
 with broadcast_imports:
     from gpaw.calculator import GPAW as OldGPAW
-    from gpaw.mixer import Mixer, MixerSum, MixerDif, MixerSum2
+    from gpaw.mixer import Mixer, MixerSum, MixerDif, MixerSum2, \
+        MixerFull
     from gpaw.eigensolvers import Davidson, RMMDIIS, CG, DirectLCAO
     from gpaw.poisson import PoissonSolver
     from gpaw.occupations import (FermiDirac, MethfesselPaxton,
