@@ -728,7 +728,7 @@ class Generator(AllElectron):
             return [divide_by_r(x_g, l) for x_g, l in zip(x_jg, vl_j)]
 
         setup.l_j = vl_j
-        setup.l_orb_j = vl_j
+        setup.l_orb_J = vl_j
         setup.n_j = vn_j
         setup.f_j = vf_j
         setup.eps_j = ve_j
