@@ -329,6 +329,7 @@ class PAWSetupGenerator:
             self.core_hole = None
 
         if projectors[-1].isupper():
+            assert projectors[-2] == ',', projectors
             self.l0 = 'SPDFG'.find(projectors[-1])
             projectors = projectors[:-2]
         else:
@@ -874,19 +875,21 @@ class PAWSetupGenerator:
             txt += '  cutoff: %.3f Bohr (tail-norm=%f)\n' % (rc, splitnorm)
 
         # Polarization:
-        gcpol = rgd.round(rcpol)
-        alpha = 1 / (0.25 * rcpol)**2
+        if lpol < 4:
+            gcpol = rgd.round(rcpol)
+            alpha = 1 / (0.25 * rcpol)**2
 
-        # Gaussian that is continuous and has a continuous derivative at rcpol:
-        phit_g = np.exp(-alpha * rgd.r_g**2) * rgd.r_g**lpol
-        phit_g -= rgd.pseudize(phit_g, gcpol, lpol, 2)[0]
-        phit_g[gcpol:] = 0.0
+            # Gaussian that is continuous and has a continuous
+            # derivative at rcpol:
+            phit_g = np.exp(-alpha * rgd.r_g**2) * rgd.r_g**lpol
+            phit_g -= rgd.pseudize(phit_g, gcpol, lpol, 2)[0]
+            phit_g[gcpol:] = 0.0
 
-        bf = BasisFunction(None, lpol, rcpol, phit_g, 'polarization')
-        self.basis.append(bf)
-        txt += 'l=%d polarization functions:\n' % lpol
-        txt += '  cutoff: %.3f Bohr (r^%d exp(-%.3f*r^2))\n' % (rcpol, lpol,
-                                                                alpha)
+            bf = BasisFunction(None, lpol, rcpol, phit_g, 'polarization')
+            self.basis.append(bf)
+            txt += f'l={lpol} polarization functions:\n'
+            txt += f'  cutoff: {rcpol:.3f} Bohr '
+            txt += f'(r^{lpol} exp(-{alpha:.3f}*r^2))\n'
 
         self.log(txt)
 
