@@ -8,13 +8,15 @@ from gpaw.mpi import MPIComm, serial_comm
 from gpaw.typing import Array1D, ArrayLike1D
 from gpaw.core.matrix import Matrix
 from gpaw.new import prod
+from gpaw.gpu import cupy as cp
 
 
 class AtomArraysLayout:
     def __init__(self,
                  shapes: list[int | tuple[int, ...]],
                  atomdist: AtomDistribution | MPIComm = serial_comm,
-                 dtype=float):
+                 dtype=float,
+                 xp=None):
         """Description of layout of atom arrays.
 
         Parameters
@@ -32,6 +34,7 @@ class AtomArraysLayout:
             atomdist = AtomDistribution(np.zeros(len(shapes), int), atomdist)
         self.atomdist = atomdist
         self.dtype = np.dtype(dtype)
+        self.xp = xp or np
 
         self.size = sum(prod(shape) for shape in self.shape_a)
 
@@ -46,12 +49,14 @@ class AtomArraysLayout:
 
     def __repr__(self):
         return (f'AtomArraysLayout({self.shape_a}, {self.atomdist}, '
-                f'{self.dtype})')
+                f'{self.dtype}, xp={self.xp.__name__})')
 
-    def new(self, atomdist=None, dtype=None):
+    def new(self, atomdist=None, dtype=None, xp=None):
         """Create new AtomsArrayLayout object with new atomdist."""
-        return AtomArraysLayout(self.shape_a, atomdist or self.atomdist,
-                                dtype or self.dtype)
+        return AtomArraysLayout(self.shape_a,
+                                atomdist or self.atomdist,
+                                dtype or self.dtype,
+                                xp or self.xp)
 
     def empty(self,
               dims: int | tuple[int, ...] = (),
@@ -67,6 +72,13 @@ class AtomArraysLayout:
         """
         return AtomArrays(self, dims, comm)
 
+    def zeros(self,
+              dims: int | tuple[int, ...] = (),
+              comm: MPIComm = serial_comm) -> AtomArrays:
+        aa = self.empty(dims, comm)
+        aa.data[:] = 0.0
+        return aa
+
     def sizes(self) -> tuple[list[dict[int, int]], Array1D]:
         """Compute array sizes for all ranks.
 
@@ -194,7 +206,7 @@ class AtomArrays:
                 raise ValueError(
                     f'Bad dtype for data: {data.dtype} != {dtype}')
         else:
-            data = np.empty(fullshape, dtype)
+            data = layout.xp.empty(fullshape, dtype)
 
         self.data = data
         self._matrix: Matrix | None = None
@@ -226,7 +238,7 @@ class AtomArrays:
 
         return self._matrix
 
-    def new(self, layout=None, data=None):
+    def new(self, *, layout=None, data=None, xp=None):
         """Create new AtomArrays object of same kind.
 
         Parameters
@@ -236,11 +248,31 @@ class AtomArrays:
         data:
             Array to use for storage.
         """
+        if xp is np:
+            assert layout is None
+            assert data is None
+            assert self.layout.xp is cp
+            layout = self.layout.new(xp=np)
         return AtomArrays(layout or self.layout,
                           self.dims,
                           self.comm,
                           data=data)
 
+    def to_cpu(self):
+        if self.layout.xp is np:
+            return self
+        return self.new(layout=self.layout.new(xp=np),
+                        data=cp.asnumpy(self.data))
+
+    def to_xp(self, xp):
+        if self.layout.xp is xp:
+            return self
+        if xp is np:
+            return self.new(layout=self.layout.new(xp=np),
+                            data=cp.asnumpy(self.data))
+        return self.new(layout=self.layout.new(xp=cp),
+                        data=cp.asarray(self.data))
+
     def __getitem__(self, a):
         if isinstance(a, numbers.Integral):
             return self._arrays[a]
@@ -285,7 +317,7 @@ class AtomArrays:
         if comm.rank == 0:
             size_ra, size_r = self.layout.sizes()
             shape = self.mydims + (size_r.max(),)
-            buffer = np.empty(shape, self.layout.dtype)
+            buffer = self.layout.xp.empty(shape, self.layout.dtype)
             for rank in range(1, comm.size):
                 buf = buffer[..., :size_r[rank]]
                 comm.receive(buf, rank)
@@ -294,6 +326,7 @@ class AtomArrays:
                     b2 = b1 + size
                     A = aa[a]
                     A[:] = buf[..., b1:b2].reshape(A.shape)
+                    b1 = b2
             for a, array in self._arrays.items():
                 aa[a] = array
         else:
@@ -352,13 +385,15 @@ class AtomArrays:
         for i1, i2 in self.layout.shape_a:
             assert i1 == i2
             shape_a.append((i1 * (i1 + 1) // 2,))
+        xp = self.layout.xp
         layout = AtomArraysLayout(shape_a,
                                   self.layout.atomdist.comm,
-                                  dtype=self.layout.dtype)
+                                  dtype=self.layout.dtype,
+                                  xp=xp)
         a_axp = layout.empty(self.dims)
         for a_xii, a_xp in zip(self.values(), a_axp.values()):
             i = a_xii.shape[-1]
-            L = np.tril_indices(i)
+            L = xp.tril_indices(i)
             for a_p, a_ii in zip(a_xp.reshape((-1, i * (i + 1) // 2)),
                                  a_xii.reshape((-1, i, i))):
                 a_p[:] = a_ii[L]
