@@ -3,9 +3,9 @@ from __future__ import annotations
 from math import pi
 
 import _gpaw
-import numpy as np
-
 import gpaw.fftw as fftw
+import numpy as np
+from ase.units import Ha
 from gpaw.core.arrays import DistributedArrays
 from gpaw.core.domain import Domain
 from gpaw.core.matrix import Matrix
@@ -19,6 +19,8 @@ from gpaw.typing import (Array1D, Array2D, Array3D, ArrayLike1D, ArrayLike2D,
 
 
 class PlaneWaves(Domain):
+    itemsize = 16
+
     def __init__(self,
                  *,
                  ecut: float,
@@ -67,7 +69,9 @@ class PlaneWaves(Domain):
         self.shape = (ng,)
         self.myshape = (len(self.ekin_G),)
 
-        self.dv = abs(np.linalg.det(self.cell_cv))
+        # Convert from np.float64 to float to avoid fake cupy problem ...
+        # XXX Fix cpupy!!!
+        self.dv = float(abs(np.linalg.det(self.cell_cv)))
 
         self._indices_cache: dict[tuple[int, ...], Array1D] = {}
 
@@ -84,6 +88,10 @@ class PlaneWaves(Domain):
         q = self.cell_cv @ self.qspiral_v / (2 * pi)
         return f'{r[:-1]}, qsiral={q}'
 
+    def _short_string(self, global_shape):
+        return (f'plane wave coefficients: {global_shape[-1]}\n'
+                f'cutoff: {self.ecut * Ha} eV\n')
+
     def global_shape(self) -> tuple[int, ...]:
         """Tuple with one element: number of plane waves."""
         return self.shape
@@ -105,7 +113,8 @@ class PlaneWaves(Domain):
 
     def empty(self,
               dims: int | tuple[int, ...] = (),
-              comm: MPIComm = serial_comm) -> PlaneWaveExpansions:
+              comm: MPIComm = serial_comm,
+              xp=None) -> PlaneWaveExpansions:
         """Create new PlaneWaveExpanions object.
 
         parameters
@@ -115,7 +124,7 @@ class PlaneWaves(Domain):
         comm:
             Distribute dimensions along this communicator.
         """
-        return PlaneWaveExpansions(self, dims, comm)
+        return PlaneWaveExpansions(self, dims, comm, xp=xp)
 
     def new(self,
             *,
@@ -196,11 +205,13 @@ class PlaneWaves(Domain):
                                 *,
                                 atomdist=None,
                                 integral=None,
-                                cut=False):
+                                cut=False,
+                                xp=None):
         """Create PlaneWaveAtomCenteredFunctions object."""
         if self.qspiral_v is None:
             return PlaneWaveAtomCenteredFunctions(functions, positions, self,
-                                                  atomdist=atomdist)
+                                                  atomdist=atomdist,
+                                                  xp=xp)
 
         from gpaw.new.spinspiral import SpiralPWACF
         return SpiralPWACF(functions, positions, self,
@@ -213,7 +224,8 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
                  pw: PlaneWaves,
                  dims: int | tuple[int, ...] = (),
                  comm: MPIComm = serial_comm,
-                 data: np.ndarray = None):
+                 data: np.ndarray = None,
+                 xp=None):
         """Object for storing function(s) as a plane-wave expansions.
 
         parameters
@@ -229,7 +241,7 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
         """
         DistributedArrays. __init__(self, dims, pw.myshape,
                                     comm, pw.comm,
-                                    data, pw.dv, complex)
+                                    data, pw.dv, complex, xp)
         self.desc = pw
         self._matrix: Matrix | None
 
@@ -237,9 +249,11 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
         txt = f'PlaneWaveExpansions(pw={self.desc}, dims={self.dims}'
         if self.comm.size > 1:
             txt += f', comm={self.comm.rank}/{self.comm.size}'
+        if self.xp is not np:
+            txt += ', xp=cp'
         return txt + ')'
 
-    def __getitem__(self, index: int) -> PlaneWaveExpansions:
+    def __getitem__(self, index: int | slice) -> PlaneWaveExpansions:
         data = self.data[index]
         return PlaneWaveExpansions(self.desc, data.shape[:-1], data=data)
 
@@ -256,7 +270,7 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
             Array to use for storage.
         """
         if data is None:
-            data = np.empty_like(self.data)
+            data = self.xp.empty_like(self.data)
         else:
             # Number of plane-waves depends on the k-point.  We therfore
             # allow for data to be bigger than needed:
@@ -292,7 +306,7 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
         return self._matrix
 
     def ifft(self, *, plan=None, grid=None, out=None, periodic=False):
-        """Do inverse FFT to uniform grid.
+        """Do inverse FFT(s) to uniform grid(s).
 
         Parameters
         ----------
@@ -304,29 +318,21 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
             Target UniformGridFunctions object.
         """
         comm = self.desc.comm
+        xp = self.xp
         if out is None:
-            out = grid.empty(self.dims)
-        assert self.desc.dtype == out.desc.dtype
+            out = grid.empty(self.dims, xp=xp)
+        assert self.desc.dtype == out.desc.dtype, (self.desc, out.desc)
         assert out.desc.pbc_c.all()
         assert comm.size == out.desc.comm.size
 
+        plan = plan or out.desc.fft_plans(xp=xp)
         this = self.gather()
         if this is not None:
-            plan = plan or out.desc.fft_plans()
             for coef_G, out1 in zip(this._arrays(), out.flat()):
-                self.desc.paste(coef_G, plan.tmp_Q)
-                if self.desc.dtype == float:
-                    t = plan.tmp_Q[:, :, 0]
-                    n, m = (s // 2 - 1 for s in out.desc.size_c[:2])
-                    t[0, -m:] = t[0, m:0:-1].conj()
-                    t[n:0:-1, -m:] = t[-n:, m:0:-1].conj()
-                    t[-n:, -m:] = t[n:0:-1, m:0:-1].conj()
-                    t[-n:, 0] = t[n:0:-1, 0].conj()
-                plan.ifft()
-                out1.scatter_from(plan.tmp_R)
+                plan.ifft_sphere(coef_G, self.desc, out1)
         else:
             for out1 in out.flat():
-                out1.scatter_from(None)
+                plan.ifft_sphere(None, self.desc, out1)
 
         if not periodic:
             out.multiply_by_eikr()
@@ -354,12 +360,12 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
         if out is None:
             if comm.rank == 0 or broadcast:
                 pw = self.desc.new(comm=serial_comm)
-                out = pw.empty(self.dims)
+                out = pw.empty(self.dims, xp=self.xp)
             else:
                 out = Empty(self.dims)
 
         if comm.rank == 0:
-            data = np.empty(self.desc.maxmysize * comm.size, complex)
+            data = self.xp.empty(self.desc.maxmysize * comm.size, complex)
         else:
             data = None
 
@@ -374,11 +380,38 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
 
         return out if not isinstance(out, Empty) else None
 
+    def gather_all(self, out: PlaneWaveExpansions) -> None:
+        """Gather coefficients from self[r] on rank r.
+
+        On rank r, an array of all G-vector coefficients will be returned.
+        These will be gathered from self[r] on all the cores.
+        """
+        assert len(self.dims) == 1
+        pw = self.desc
+        comm = pw.comm
+        if comm.size == 1:
+            out.data[:] = self.data[0]
+            return
+
+        N = self.dims[0]
+        assert N <= comm.size
+
+        ng = pw.shape[0]
+        myng = pw.myshape[0]
+        maxmyng = pw.maxmysize
+
+        ssize_r, soffset_r, rsize_r, roffset_r = a2a_stuff(
+            comm, N, ng, myng, maxmyng)
+
+        comm.alltoallv(self.data, ssize_r, soffset_r,
+                       out.data, rsize_r, roffset_r)
+
     def scatter_from(self, data: Array1D = None) -> None:
         """Scatter data from rank-0 to all ranks."""
         comm = self.desc.comm
         if comm.size == 1:
-            self.data[:] = data
+            assert data is not None
+            self.data[:] = self.xp.asarray(data)
             return
 
         assert self.dims == ()
@@ -387,37 +420,60 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
             data = pad(data, comm.size * self.desc.maxmysize)
             comm.scatter(data, self.data, 0)
         else:
-            buf = np.empty(self.desc.maxmysize, complex)
+            buf = self.xp.empty(self.desc.maxmysize, complex)
             comm.scatter(None, buf, 0)
             self.data[:] = buf[:len(self.data)]
 
+    def scatter_from_all(self, a_G: PlaneWaveExpansions) -> None:
+        """Scatter all coefficients from rank r to self on other cores."""
+        assert len(self.dims) == 1
+        pw = self.desc
+        comm = pw.comm
+        if comm.size == 1:
+            self.data[:] = a_G.data
+            return
+
+        N = self.dims[0]
+        assert N <= comm.size
+
+        ng = pw.shape[0]
+        myng = pw.myshape[0]
+        maxmyng = pw.maxmysize
+
+        rsize_r, roffset_r, ssize_r, soffset_r = a2a_stuff(
+            comm, N, ng, myng, maxmyng)
+
+        comm.alltoallv(a_G.data, ssize_r, soffset_r,
+                       self.data, rsize_r, roffset_r)
+
     def integrate(self, other: PlaneWaveExpansions = None) -> np.ndarray:
         """Integral of self or self time cc(other)."""
+        dv = self.dv
         if other is not None:
             assert self.comm.size == 1
             assert self.desc.dtype == other.desc.dtype
             a = self._arrays()
             b = other._arrays()
-            dv = self.dv
             if self.desc.dtype == float:
                 a = a.view(float)
                 b = b.view(float)
                 dv *= 2
             result = a @ b.T.conj()
             if self.desc.dtype == float and self.desc.comm.rank == 0:
-                result -= 0.5 * np.outer(a[:, 0], b[:, 0])
+                result -= 0.5 * a[:, :1] @ b[:, :1].T
             self.desc.comm.sum(result)
-            result.shape = self.dims + other.dims
+            result = result.reshape(self.dims + other.dims)
         else:
-            dv = self.dv
             if self.desc.comm.rank == 0:
                 result = self.data[..., 0]
             else:
-                result = np.empty(self.mydims, complex)
+                result = self.xp.empty(self.mydims, complex)
             self.desc.comm.broadcast(result, 0)
 
         if self.desc.dtype == float:
             result = result.real
+        if result.ndim == 0:
+            result = result.item()  # convert to scalar
         return result * dv
 
     def _matrix_elements_correction(self,
@@ -428,13 +484,13 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
         if self.desc.dtype == float:
             out.data *= 2.0
             if self.desc.comm.rank == 0:
-                correction = np.outer(M1.data[:, 0],
-                                      M2.data[:, 0]) * self.dv
+                correction = M1.data[:, :1] @ M2.data[:, :1].T
                 if symmetric:
-                    correction *= 0.5
+                    correction *= 0.5 * self.dv
                     out.data -= correction
                     out.data -= correction.T
                 else:
+                    correction *= self.dv
                     out.data -= correction
 
     def norm2(self, kind: str = 'normal') -> np.ndarray:
@@ -459,11 +515,14 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
         """
         a_xG = self._arrays().view(float)
         if kind == 'normal':
-            result_x = np.einsum('xG, xG -> x', a_xG, a_xG)
+            result_x = self.xp.einsum('xG, xG -> x', a_xG, a_xG)
         elif kind == 'kinetic':
-            a_xG.shape = (len(a_xG), -1, 2)
-            result_x = np.einsum('xGi, xGi, G -> x',
-                                 a_xG, a_xG, self.desc.ekin_G)
+
+            a_xG = a_xG.reshape((len(a_xG), -1, 2))
+            result_x = self.xp.einsum('xGi, xGi, G -> x',
+                                      a_xG,
+                                      a_xG,
+                                      self.xp.asarray(self.desc.ekin_G))
         else:
             1 / 0
         if self.desc.dtype == float:
@@ -471,20 +530,63 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
             if self.desc.comm.rank == 0 and kind == 'normal':
                 result_x -= a_xG[:, 0]**2
         self.desc.comm.sum(result_x)
-        result_x.shape = self.mydims
-        return result_x * self.dv
+        return result_x.reshape(self.mydims) * self.dv
 
     def abs_square(self,
                    weights: Array1D,
-                   out: UniformGridFunctions = None) -> None:
-        """Add weighted absolute square of data to output array."""
-        assert out is not None
-        tmp_R = out.desc.new(dtype=self.desc.dtype).empty()
-        for f, psit_G in zip(weights, self):
-            # Same as (but much faster):
-            # out.data += f * abs(psit.ifft().data)**2
-            psit_G.ifft(out=tmp_R)
-            _gpaw.add_to_density(f, tmp_R.data, out.data)
+                   out: UniformGridFunctions) -> None:
+        """Add weighted absolute square of self to output array.
+
+        With `a_n(G)` being self and `w_n` the weights:::
+
+              _         _    --     -1    _   2
+          out(r) <- out(r) + >  |FFT  [a (G)]| w
+                             --         n       n
+                             n
+
+        """
+        pw = self.desc
+        domain_comm = pw.comm
+        xp = self.xp
+        a_nG = self
+
+        if domain_comm.size == 1:
+            a_R = out.desc.new(dtype=pw.dtype).empty(xp=xp)
+            for weight, a_G in zip(weights, a_nG):
+                if weight == 0.0:
+                    continue
+                a_G.ifft(out=a_R)
+                if xp is np:
+                    _gpaw.add_to_density(weight, a_R.data, out.data)
+                else:
+                    out.data += float(weight) * xp.abs(a_R.data)**2
+            return
+
+        # Undistributed work arrays:
+        a1_R = out.desc.new(comm=None, dtype=pw.dtype).empty(xp=xp)
+        a1_G = pw.new(comm=None).empty(xp=xp)
+        b1_R = out.desc.new(comm=None).zeros(xp=xp)
+
+        (N,) = self.mydims
+        for n1 in range(0, N, domain_comm.size):
+            n2 = min(n1 + domain_comm.size, N)
+            a_nG[n1:n2].gather_all(a1_G)
+            n = n1 + domain_comm.rank
+            if n >= N:
+                continue
+            weight = weights[n]
+            if weight == 0.0:
+                continue
+            a1_G.ifft(out=a1_R)
+            if xp is np:
+                _gpaw.add_to_density(weight, a1_R.data, b1_R.data)
+            else:
+                b1_R.data += float(weight) * xp.abs(a1_R.data)**2
+
+        domain_comm.sum(b1_R.data)
+        b_R = out.new()
+        b_R.scatter_from(b1_R)
+        out.data += b_R.data
 
     def to_pbc_grid(self):
         return self
@@ -492,11 +594,63 @@ class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
     def randomize(self) -> None:
         """Insert random numbers between -0.5 and 0.5 into data."""
         seed = [self.comm.rank, self.desc.comm.rank]
-        rng = np.random.default_rng(seed)
+        rng = self.xp.random.default_rng(seed)
         a = self.data.view(float)
         rng.random(a.shape, out=a)
         a -= 0.5
 
+    def moment(self):
+        pw = self.desc
+        # Masks:
+        m0_G, m1_G, m2_G = [i_G == 0 for i_G in pw.indices_cG]
+        a_G = self.gather()
+        if a_G is not None:
+            b_G = a_G.data.imag
+            b_cs = [b_G[m1_G & m2_G],
+                    b_G[m0_G & m2_G],
+                    b_G[m0_G & m1_G]]
+            d_c = [b_s[1:] @ (1.0 / np.arange(1, len(b_s)))
+                   for b_s in b_cs]
+            m_v = np.dot(d_c, pw.cell_cv) / pi * pw.dv
+        else:
+            m_v = np.empty(3)
+        pw.comm.broadcast(m_v, 0)
+        return m_v
+
+    def morph(self, pw):
+        pw0 = self.desc
+        out_xG = pw.zeros(self.dims,
+                          comm=self.comm,
+                          xp=self.xp)
+
+        d = {}
+        for G, i_c in enumerate(pw.indices_cG.T):
+            d[tuple(i_c)] = G
+        G_G0 = []
+        G0_G = []
+        for G0, i_c in enumerate(pw0.indices_cG.T):
+            G = d.get(tuple(i_c))
+            if G is not None:
+                G_G0.append(G)
+                G0_G.append(G0)
+
+        out_xG.data[:, G_G0] = self.data[:, G0_G]
+        return out_xG
+
+
+def a2a_stuff(comm, N, ng, myng, maxmyng):
+    """Create arrays for MPI alltoallv call."""
+    ssize_r = np.zeros(comm.size, int)
+    ssize_r[:N] = myng
+    soffset_r = np.arange(comm.size) * myng
+    soffset_r[N:] = 0
+    roffset_r = (np.arange(comm.size) * maxmyng).clip(max=ng)
+    rsize_r = np.zeros(comm.size, int)
+    if comm.rank < N:
+        rsize_r[:-1] = roffset_r[1:] - roffset_r[:-1]
+        rsize_r[-1] = ng - roffset_r[-1]
+    return ssize_r, soffset_r, rsize_r, roffset_r
+
 
 class Empty:
     def __init__(self, dims):
@@ -576,71 +730,3 @@ def find_reciprocal_vectors(ecut: float,
     G_plus_k = G_plus_k_Qv[mask]
 
     return G_plus_k, ekin, indices.T
-
-
-x = '''
-class PWMapping:
-    def __init__(self, pw1: PlaneWaves, pw2: PlaneWaves):
-        """Mapping from pd1 to pd2."""
-        N_c = pw1.grid.size
-        N2_c = pw2.grid.size
-        assert pw1.grid.dtype == pw2.grid.dtype
-        if pw1.grid.dtype == float:
-            N_c = N_c.copy()
-            N_c[2] = N_c[2] // 2 + 1
-            N2_c = N2_c.copy()
-            N2_c[2] = N2_c[2] // 2 + 1
-
-        Q1_G = pw1.myindices
-        Q1_Gc = np.empty((len(Q1_G), 3), int)
-        Q1_Gc[:, 0], r_G = divmod(Q1_G, N_c[1] * N_c[2])
-        Q1_Gc.T[1:] = divmod(r_G, N_c[2])
-        if pw1.grid.dtype == float:
-            C = 2
-        else:
-            C = 3
-        Q1_Gc[:, :C] += N_c[:C] // 2
-        Q1_Gc[:, :C] %= N_c[:C]
-        Q1_Gc[:, :C] -= N_c[:C] // 2
-        Q1_Gc[:, :C] %= N2_c[:C]
-        Q2_G = Q1_Gc[:, 2] + N2_c[2] * (Q1_Gc[:, 1] + N2_c[1] * Q1_Gc[:, 0])
-        G2_Q = np.empty(N2_c, int).ravel()
-        G2_Q[:] = -1
-        G2_Q[pw2.myindices] = np.arange(len(pw2.myindices))
-        G2_G1 = G2_Q[Q2_G]
-
-        if pw1.grid.comm.size == 1:
-            self.G2_G1 = G2_G1
-            self.G1 = None
-        else:
-            mask_G1 = (G2_G1 != -1)
-            self.G2_G1 = G2_G1[mask_G1]
-            self.G1 = np.arange(pw1.maxmysize)[mask_G1]
-
-        self.pw1 = pw1
-        self.pw2 = pw2
-
-    def add_to1(self, a_G1, b_G2):
-        """Do a += b * scale, where a is on pd1 and b on pd2."""
-        scale = self.pd1.tmp_R.size / self.pd2.tmp_R.size
-
-        if self.pd1.gd.comm.size == 1:
-            a_G1 += b_G2[self.G2_G1] * scale
-            return
-
-        b_G1 = self.pd1.tmp_G
-        b_G1[:] = 0.0
-        b_G1[self.G1] = b_G2[self.G2_G1]
-        self.pd1.gd.comm.sum(b_G1)
-        ng1 = self.pd1.gd.comm.rank * self.pd1.maxmyng
-        ng2 = ng1 + self.pd1.myng_q[0]
-        a_G1 += b_G1[ng1:ng2] * scale
-
-    def add_to2(self, a2, b1):
-        """Do a += b * scale, where a is on pd2 and b on pd1."""
-        myb = b1.data * (self.pw2.grid.shape[0] / self.pw1.grid.shape[0])
-        if self.desc1.grid.comm.size == 1:
-            a2.data[self.G2_G1] += myb
-        else:
-            1 / 0
-'''
