@@ -1,15 +1,21 @@
+from __future__ import annotations
 from math import pi
 
 import _gpaw
+import gpaw.gpu.kernels as gpu_kernels
 import numpy as np
 from gpaw.core.atom_arrays import AtomArraysLayout, AtomDistribution
 from gpaw.core.atom_centered_functions import AtomCenteredFunctions
+from gpaw.core.uniform_grid import UniformGridFunctions
+from gpaw.gpu import cupy_is_fake
 from gpaw.lfc import BaseLFC
+from gpaw.new import prod
 from gpaw.pw.lfc import ft
 from gpaw.spherical_harmonics import Y, nablarlYL
 from gpaw.utilities.blas import mmm
-from gpaw.core.uniform_grid import UniformGridFunctions
-from gpaw.new import prod
+from typing import TYPE_CHECKING
+if TYPE_CHECKING:
+    from gpaw.core.plane_waves import PlaneWaves
 
 
 class PlaneWaveAtomCenteredFunctions(AtomCenteredFunctions):
@@ -17,15 +23,17 @@ class PlaneWaveAtomCenteredFunctions(AtomCenteredFunctions):
                  functions,
                  fracpos,
                  pw,
-                 atomdist=None):
+                 atomdist=None,
+                 xp=None):
         AtomCenteredFunctions.__init__(self, functions, fracpos, atomdist)
         self.pw = pw
+        self.xp = xp or np
 
     def _lazy_init(self):
         if self._lfc is not None:
             return
 
-        self._lfc = PWLFC(self.functions, self.pw)
+        self._lfc = PWLFC(self.functions, self.pw, xp=self.xp)
 
         if self._atomdist is None:
             self._atomdist = AtomDistribution.from_number_of_atoms(
@@ -37,18 +45,28 @@ class PlaneWaveAtomCenteredFunctions(AtomCenteredFunctions):
         self._layout = AtomArraysLayout([sum(2 * f.l + 1 for f in funcs)
                                          for funcs in self.functions],
                                         self._atomdist,
-                                        self.pw.dtype)
+                                        self.pw.dtype,
+                                        xp=self.xp)
+
+    def __repr__(self):
+        s = super().__repr__()
+        if self.xp is np:
+            return s
+        return s[:-1] + ', xp=cp)'
 
     def to_uniform_grid(self,
                         out: UniformGridFunctions,
                         scale: float = 1.0) -> UniformGridFunctions:
-        out_G = self.pw.zeros()
+        out_G = self.pw.zeros(xp=out.xp)
         self.add_to(out_G, scale)
         return out_G.ifft(out=out)
 
 
 class PWLFC(BaseLFC):
-    def __init__(self, functions, pw, blocksize=5000):
+    def __init__(self,
+                 functions,
+                 pw: PlaneWaves,
+                 blocksize=5000, *, xp):
         """Reciprocal-space plane-wave localized function collection.
 
         spline_aj: list of list of spline objects
@@ -60,6 +78,7 @@ class PWLFC(BaseLFC):
             doing all G-vectors in one big block.
         """
 
+        self.xp = xp
         self.pw = pw
         self.spline_aj = functions
 
@@ -96,6 +115,8 @@ class PWLFC(BaseLFC):
         if self.initialized:
             return
 
+        xp = self.xp
+
         splines = {}  # Dict[Spline, int]
         for spline_j in self.spline_aj:
             for spline in spline_j:
@@ -105,34 +126,42 @@ class PWLFC(BaseLFC):
 
         nJ = sum(len(spline_j) for spline_j in self.spline_aj)
 
-        self.f_Gs = np.empty(self.pw.myshape + (nsplines,))
+        self.f_Gs = xp.empty(self.pw.myshape + (nsplines,))
         self.l_s = np.empty(nsplines, np.int32)
         self.a_J = np.empty(nJ, np.int32)
         self.s_J = np.empty(nJ, np.int32)
-
+        self.I_J = np.empty(nJ, np.int32)
         # Fourier transform radial functions:
         J = 0
         done = set()  # Set[Spline]
+        I = 0
         for a, spline_j in enumerate(self.spline_aj):
             for spline in spline_j:
                 s = splines[spline]  # get spline index
                 if spline not in done:
                     f = ft(spline)
                     G_G = (2 * self.pw.ekin_G)**0.5
-                    self.f_Gs[:, s] = f.map(G_G)
+                    self.f_Gs[:, s] = xp.asarray(f.map(G_G))
                     self.l_s[s] = spline.get_angular_momentum_number()
                     done.add(spline)
                 self.a_J[J] = a
                 self.s_J[J] = s
+                self.I_J[J] = I
+                I += 2 * spline.get_angular_momentum_number() + 1
                 J += 1
 
         self.lmax = max(self.l_s, default=-1)
 
         # Spherical harmonics:
         G_Gv = self.pw.G_plus_k_Gv
-        self.Y_GL = np.empty((len(G_Gv), (self.lmax + 1)**2))
+        self.Y_GL = xp.empty((len(G_Gv), (self.lmax + 1)**2))
         for L in range((self.lmax + 1)**2):
-            self.Y_GL[:, L] = Y(L, *G_Gv.T)
+            self.Y_GL[:, L] = xp.asarray(Y(L, *G_Gv.T))
+
+        self.l_s = xp.asarray(self.l_s)
+        self.a_J = xp.asarray(self.a_J)
+        self.s_J = xp.asarray(self.s_J)
+        self.I_J = xp.asarray(self.I_J)
 
         self.initialized = True
 
@@ -143,16 +172,19 @@ class PWLFC(BaseLFC):
     def set_positions(self, spos_ac, atomdist):
         self.initialize()
 
+        xp = self.xp
+
         if self.pw.dtype == float:
-            self.eikR_a = np.ones(len(spos_ac))
+            self.eikR_a = xp.ones(len(spos_ac))
         else:
-            self.eikR_a = np.exp(2j * pi * (spos_ac @ self.pw.kpt_c))
+            self.eikR_a = xp.asarray(
+                np.exp(2j * pi * (spos_ac @ self.pw.kpt_c)))
 
         self.pos_av = np.dot(spos_ac, self.pw.cell)
 
-        Gk_Gv = self.pw.G_plus_k_Gv
-        GkR_Ga = Gk_Gv @ self.pos_av.T
-        self.emiGR_Ga = np.exp(-1j * GkR_Ga) * self.eikR_a
+        Gk_Gv = xp.asarray(self.pw.G_plus_k_Gv)
+        GkR_Ga = Gk_Gv @ xp.asarray(self.pos_av.T)
+        self.emiGR_Ga = xp.exp(-1j * GkR_Ga) * self.eikR_a
 
         rank_a = atomdist.rank_a
 
@@ -179,6 +211,8 @@ class PWLFC(BaseLFC):
         cc: bool
             Complex conjugate.
         """
+        xp = self.xp
+
         if G2 is None:
             G2 = self.Y_GL.shape[0]
 
@@ -187,7 +221,7 @@ class PWLFC(BaseLFC):
         Y_GL = self.Y_GL[G1:G2]
 
         if self.dtype == complex:
-            f_GI = np.empty((G2 - G1, self.nI), complex)
+            f_GI = xp.empty((G2 - G1, self.nI), complex)
         else:
             # Special layout because BLAS does not have real-complex
             # multiplications.  f_GI(G,I) layout:
@@ -198,17 +232,22 @@ class PWLFC(BaseLFC):
             #    imag(G1+1, 0), imag(G1+1, 1), ...
             #    ...
 
-            f_GI = np.empty((2 * (G2 - G1), self.nI))
+            f_GI = xp.empty((2 * (G2 - G1), self.nI))
 
-        if True:
+        if xp is np:
             # Fast C-code:
             _gpaw.pwlfc_expand(f_Gs, emiGR_Ga, Y_GL,
                                self.l_s, self.a_J, self.s_J,
                                cc, f_GI)
             return f_GI
+        elif cupy_is_fake or getattr(_gpaw, 'gpu_aware_mpi', False):
+            gpu_kernels.pwacf_expand(f_Gs, emiGR_Ga, Y_GL,
+                                     self.l_s, self.a_J, self.s_J,
+                                     cc, f_GI, self.I_J)
+            return f_GI
 
         # Equivalent slow Python code:
-        f_GI = np.empty((G2 - G1, self.nI), complex)
+        f_GI = xp.empty((G2 - G1, self.nI), complex)
         I1 = 0
         for J, (a, s) in enumerate(zip(self.a_J, self.s_J)):
             l = self.l_s[s]
@@ -220,7 +259,7 @@ class PWLFC(BaseLFC):
             I1 = I2
         if cc:
             f_GI = f_GI.conj()
-        if self.pd.dtype == float:
+        if self.dtype == float:
             f_GI = f_GI.T.copy().view(float).T.copy()
 
         return f_GI
@@ -236,7 +275,7 @@ class PWLFC(BaseLFC):
                 G1 = G2
             if ensure_same_number_of_blocks:
                 # Make sure we yield the same number of times:
-                nb = (self.pd.maxmyng + B - 1) // B
+                nb = (self.pw.maxmysize + B - 1) // B
                 mynb = (nG + B - 1) // B
                 if mynb < nb:
                     yield nG, nG  # empty block
@@ -244,7 +283,9 @@ class PWLFC(BaseLFC):
             yield 0, nG
 
     def add(self, a_xG, c_axi=1.0, q=None):
-        c_xI = np.empty(a_xG.shape[:-1] + (self.nI,), self.dtype)
+        if self.nI == 0:
+            return
+        c_xI = self.xp.empty(a_xG.shape[:-1] + (self.nI,), self.dtype)
 
         if isinstance(c_axi, float):
             assert a_xG.ndim == 1
@@ -269,11 +310,19 @@ class PWLFC(BaseLFC):
                 G1 *= 2
                 G2 *= 2
 
-            mmm(1.0 / self.pw.dv, c_xI, 'N', f_GI, 'T',
-                1.0, a_xG[:, G1:G2])
+            if self.xp is np:
+                mmm(1.0 / self.pw.dv, c_xI, 'N', f_GI, 'T',
+                    1.0, a_xG[:, G1:G2])
+            else:
+                self.xp.cublas.gemm('N', 'T',
+                                    c_xI, f_GI, a_xG[:, G1:G2],
+                                    1.0 / self.pw.dv, 1.0)
 
     def integrate(self, a_xG, c_axi=None, q=-1):
-        c_xI = np.zeros(a_xG.shape[:-1] + (self.nI,), self.dtype)
+        xp = self.xp
+        if self.nI == 0:
+            return c_axi
+        c_xI = xp.zeros(a_xG.shape[:-1] + (self.nI,), self.dtype)
 
         nx = prod(c_xI.shape[:-1])
         b_xI = c_xI.reshape((nx, self.nI))
@@ -295,7 +344,12 @@ class PWLFC(BaseLFC):
                     f_GI[0] *= 0.5
                 G1 *= 2
                 G2 *= 2
-            mmm(alpha, a_xG[:, G1:G2], 'N', f_GI, 'N', x, b_xI)
+            if xp is np:
+                mmm(alpha, a_xG[:, G1:G2], 'N', f_GI, 'N', x, b_xI)
+            else:
+                xp.cublas.gemm('N', 'N',
+                               a_xG[:, G1:G2], f_GI, b_xI,
+                               alpha, x)
             x = 1.0
 
         self.comm.sum(b_xI)
@@ -305,7 +359,8 @@ class PWLFC(BaseLFC):
         return c_axi
 
     def derivative(self, a_xG, c_axiv=None, q=-1):
-        c_vxI = np.zeros((3,) + a_xG.shape[:-1] + (self.nI,), self.dtype)
+        xp = self.xp
+        c_vxI = xp.zeros((3,) + a_xG.shape[:-1] + (self.nI,), self.dtype)
         nx = prod(c_vxI.shape[1:-1])
         b_vxI = c_vxI.reshape((3, nx, self.nI))
         a_xG = a_xG.reshape((nx, a_xG.shape[-1])).view(self.dtype)
@@ -318,22 +373,36 @@ class PWLFC(BaseLFC):
         x = 0.0
         for G1, G2 in self.block():
             f_GI = self.expand(G1, G2, cc=True)
-            G_Gv = self.pw.G_plus_k_Gv[G1:G2]
+            G_Gv = xp.asarray(self.pw.G_plus_k_Gv[G1:G2])
             if self.dtype == float:
-                d_GI = np.empty_like(f_GI)
+                d_GI = xp.empty(f_GI.shape)
                 for v in range(3):
                     d_GI[::2] = f_GI[1::2] * G_Gv[:, v, np.newaxis]
                     d_GI[1::2] = f_GI[::2] * G_Gv[:, v, np.newaxis]
-                    mmm(2 * alpha,
-                        a_xG[:, 2 * G1:2 * G2], 'N',
-                        d_GI, 'N',
-                        x, b_vxI[v])
+                    if xp is np:
+                        mmm(2 * alpha,
+                            a_xG[:, 2 * G1:2 * G2], 'N',
+                            d_GI, 'N',
+                            x, b_vxI[v])
+                    else:
+                        xp.cublas.gemm('N', 'N',
+                                       a_xG[:, 2 * G1:2 * G2],
+                                       d_GI,
+                                       b_vxI[v],
+                                       2 * alpha, x)
             else:
                 for v in range(3):
-                    mmm(-alpha,
-                        a_xG[:, G1:G2], 'N',
-                        f_GI * G_Gv[:, v, np.newaxis], 'N',
-                        x, b_vxI[v])
+                    if xp is np:
+                        mmm(-alpha,
+                            a_xG[:, G1:G2], 'N',
+                            f_GI * G_Gv[:, v, np.newaxis], 'N',
+                            x, b_vxI[v])
+                    else:
+                        xp.cublas.gemm('N', 'N',
+                                       a_xG[:, G1:G2],
+                                       f_GI * G_Gv[:, v, np.newaxis],
+                                       b_vxI[v],
+                                       -alpha, x)
             x = 1.0
 
         self.comm.sum(c_vxI)
@@ -349,7 +418,8 @@ class PWLFC(BaseLFC):
 
         return c_axiv
 
-    def stress_tensor_contribution(self, a_xG, c_axi=1.0, q=-1):
+    def stress_tensor_contribution(self, a_xG, c_axi=1.0):
+        xp = self.xp
         cache = {}
         things = []
         I1 = 0
@@ -358,7 +428,7 @@ class PWLFC(BaseLFC):
             for spline in spline_j:
                 if spline not in cache:
                     s = ft(spline)
-                    G_G = self.pd.G2_qG[q]**0.5
+                    G_G = (2 * self.pw.ekin_G)**0.5
                     f_G = []
                     dfdGoG_G = []
                     for G in G_G:
@@ -367,8 +437,8 @@ class PWLFC(BaseLFC):
                             G = 1.0
                         f_G.append(f)
                         dfdGoG_G.append(dfdG / G)
-                    f_G = np.array(f_G)
-                    dfdGoG_G = np.array(dfdGoG_G)
+                    f_G = xp.array(f_G)
+                    dfdGoG_G = xp.array(dfdGoG_G)
                     cache[spline] = (f_G, dfdGoG_G)
                 else:
                     f_G, dfdGoG_G = cache[spline]
@@ -381,28 +451,30 @@ class PWLFC(BaseLFC):
         if isinstance(c_axi, float):
             c_axi = dict((a, c_axi) for a in range(len(self.pos_av)))
 
-        G0_Gv = self.pd.get_reciprocal_vectors(q=q)
+        G0_Gv = self.pw.G_plus_k_Gv
 
-        stress_vv = np.zeros((3, 3))
+        stress_vv = xp.zeros((3, 3))
         for G1, G2 in self.block(ensure_same_number_of_blocks=True):
             G_Gv = G0_Gv[G1:G2]
-            Z_LvG = np.array([nablarlYL(L, G_Gv.T)
+            Z_LvG = xp.array([nablarlYL(L, G_Gv.T)
                               for L in range((lmax + 1)**2)])
+            G_Gv = xp.asarray(G_Gv)
             aa_xG = a_xG[..., G1:G2]
             for v1 in range(3):
                 for v2 in range(3):
                     stress_vv[v1, v2] += self._stress_tensor_contribution(
-                        v1, v2, things, G1, G2, G_Gv, aa_xG, c_axi, q, Z_LvG)
+                        v1, v2, things, G1, G2, G_Gv, aa_xG, c_axi, Z_LvG)
 
         self.comm.sum(stress_vv)
 
         return stress_vv
 
     def _stress_tensor_contribution(self, v1, v2, things, G1, G2,
-                                    G_Gv, a_xG, c_axi, q, Z_LvG):
-        f_IG = np.empty((self.nI, G2 - G1), complex)
-        emiGR_Ga = self.emiGR_qGa[q][G1:G2]
-        Y_LG = self.Y_qGL[q].T
+                                    G_Gv, a_xG, c_axi, Z_LvG):
+        xp = self.xp
+        f_IG = xp.empty((self.nI, G2 - G1), complex)
+        emiGR_Ga = self.emiGR_Ga[G1:G2]
+        Y_LG = self.Y_GL.T
         for a, l, I1, I2, f_G, dfdGoG_G in things:
             L1 = l**2
             L2 = (l + 1)**2
@@ -411,24 +483,27 @@ class PWLFC(BaseLFC):
                             Y_LG[L1:L2, G1:G2] +
                             f_G[G1:G2] * G_Gv[:, v1] * Z_LvG[L1:L2, v2]))
 
-        c_xI = np.zeros(a_xG.shape[:-1] + (self.nI,), self.pd.dtype)
+        c_xI = xp.zeros(a_xG.shape[:-1] + (self.nI,), self.pw.dtype)
 
         x = prod(c_xI.shape[:-1])
         b_xI = c_xI.reshape((x, self.nI))
         a_xG = a_xG.reshape((x, a_xG.shape[-1]))
 
-        alpha = 1.0 / self.pd.gd.N_c.prod()
-        if self.pd.dtype == float:
-            alpha *= 2
-            if G1 == 0 and self.pd.gd.comm.rank == 0:
+        alpha = 1.0
+        if self.pw.dtype == float:
+            alpha = 2.0
+            if G1 == 0 and self.pw.comm.rank == 0:
                 f_IG[:, 0] *= 0.5
             f_IG = f_IG.view(float)
             a_xG = a_xG.copy().view(float)
 
-        mmm(alpha, a_xG, 'N', f_IG, 'C', 0.0, b_xI)
+        if xp is np:
+            mmm(alpha, a_xG, 'N', f_IG, 'C', 0.0, b_xI)
+        else:
+            xp.cublas.gemm('N', 'H', a_xG, f_IG, b_xI, alpha, 0.0)
         self.comm.sum(b_xI)
 
         stress = 0.0
         for a, I1, I2 in self.my_indices:
-            stress -= self.eikR_qa[q][a] * (c_axi[a] * c_xI[..., I1:I2]).sum()
+            stress -= self.eikR_a[a] * (c_axi[a] * c_xI[..., I1:I2]).sum()
         return stress.real
